import json
import logging
from abc import abstractmethod
from argparse import Namespace
from functools import lru_cache
from pathlib import Path
from typing import Mapping

from datasets import load_dataset
from llama_cpp import LlamaGrammar

from evoprompt.helpers.prompts import BasePromptsFromJsonMixin
from evoprompt.task import Task, TextClassification
from evoprompt.task.task import DatasetDatum

logger = logging.getLogger(__name__)


# a simple grammar in GBNF notation which allows either positive or negative
# NOTE: for some reason the program crashes if the grammar is an instance attribute of the task class
# TODO still needed?


class SentimentAnalysis(TextClassification):
    @lru_cache
    def _get_label_mapping(self) -> Mapping:
        return {"negative": 0, "positive": 1}


class HfSST2(BasePromptsFromJsonMixin, SentimentAnalysis):
    shorthand = "hf-sst2"
    base_prompts_file = "evoprompt/initial_prompts/sst-2/prompts.json"

    def load_validation_set(
        self, validation_dataset: str | None, validation_split: str | None
    ):
        return super().load_validation_set("stanfordnlp/sst2", "validation")

    def load_test_set(self, test_dataset: str, test_split: str | None):
        return super().load_test_set("stanfordnlp/sst2", "test")

    def _get_text_for_datum(self, datum: DatasetDatum) -> str:
        return datum["sentence"]

    def _get_gold_label_for_datum(self, datum: DatasetDatum) -> str:
        return datum["label"]


class SST2(BasePromptsFromJsonMixin, SentimentAnalysis):
    shorthand = "sst2"
    base_prompts_file = "evoprompt/initial_prompts/sst-2/prompts.json"

    def load_validation_set(
        self, validation_dataset: str | None, validation_split: str | None
    ):
        return load_dataset(
            "csv",
            data_files={"validation": "evoprompt/data/sst-2/train.tsv"},
            split="validation",
            sep="\t",
        )

    def load_test_set(self, test_dataset: str, test_split: str | None):
        # this dataset is not in correct format therefore we process it after reading as a simple text file
        dataset = load_dataset(
            "text",
            data_files={"test": "evoprompt/data/sst-2/test.tsv"},
            split="test",
        )
        dataset = dataset.map(
            lambda sample: {
                "label": int(sample["text"][0]),
                "sentence": sample["text"][2:],
            },
            remove_columns="text",
        )
        return dataset

    def _get_text_for_datum(self, datum: DatasetDatum) -> str:
        return datum["sentence"]

    def _get_gold_label_for_datum(self, datum: DatasetDatum) -> str:
        return datum["label"]


class SST5(BasePromptsFromJsonMixin, SentimentAnalysis):
    shorthand = "sst5"
    base_prompts_file = "evoprompt/initial_prompts/sst-5/prompts.json"

    def load_validation_set(
        self, validation_dataset: str | None, validation_split: str | None
    ):
        return load_dataset(
            "csv",
            data_files={"validation": "evoprompt/data/sst-5/train.csv"},
            split="validation",
            column_names=["label", "sentence"],
        )

    def load_test_set(self, test_dataset: str, test_split: str | None):
        return load_dataset(
            "csv",
            data_files={"test": "evoprompt/data/sst-5/test.csv"},
            split="test",
            column_names=["label", "sentence"],
        )

    def _get_text_for_datum(self, datum: DatasetDatum) -> str:
        return datum["sentence"]

    def _get_gold_label_for_datum(self, datum: DatasetDatum) -> str:
        return datum["label"]

    @lru_cache
    def _get_label_mapping(self) -> Mapping:
        classes = ["terrible", "bad", "okay", "good", "great"]
        return dict(zip(classes, range(len(classes))))


class HfMovieReviews(BasePromptsFromJsonMixin, SentimentAnalysis):
    shorthand = "hf-mr"
    base_prompts_file = "evoprompt/initial_prompts/mr/prompts.json"

    def load_validation_set(
        self, validation_dataset: str | None, validation_split: str | None
    ):
        return super().load_validation_set(
            "cornell-movie-review-data/rotten_tomatoes", "validation"
        )

    def load_test_set(self, test_dataset: str, test_split: str | None):
        return super().load_test_set(
            "cornell-movie-review-data/rotten_tomatoes", "test"
        )

    def _get_text_for_datum(self, datum: DatasetDatum) -> str:
        return datum["text"]

    def _get_gold_label_for_datum(self, datum: DatasetDatum) -> str:
        return datum["label"]


class MovieReviews(BasePromptsFromJsonMixin, SentimentAnalysis):
    shorthand = "mr"
    base_prompts_file = "evoprompt/initial_prompts/mr/prompts.json"

    def load_validation_set(
        self, validation_dataset: str | None, validation_split: str | None
    ):
        return load_dataset(
            "csv",
            data_files={"validation": "evoprompt/data/mr/train.csv"},
            split="validation",
            column_names=["label", "text"],
        )

    def load_test_set(self, test_dataset: str, test_split: str | None):
        return load_dataset(
            "csv",
            data_files={"test": "evoprompt/data/mr/test.csv"},
            split="test",
            column_names=["label", "text"],
        )

    def _get_text_for_datum(self, datum: DatasetDatum) -> str:
        return datum["text"]

    def _get_gold_label_for_datum(self, datum: DatasetDatum) -> str:
        return datum["label"]


class CustomerReviews(BasePromptsFromJsonMixin, SentimentAnalysis):
    shorthand = "cr"
    base_prompts_file = "evoprompt/initial_prompts/cr/prompts.json"

    def load_validation_set(
        self, validation_dataset: str | None, validation_split: str | None
    ):
        return load_dataset(
            "csv",
            data_files={"validation": "evoprompt/data/cr/train.csv"},
            split="validation",
            column_names=["label", "text"],
        )

    def load_test_set(self, test_dataset: str, test_split: str | None):
        return load_dataset(
            "csv",
            data_files={"test": "evoprompt/data/cr/test.csv"},
            split="test",
            column_names=["label", "text"],
        )

    def _get_text_for_datum(self, datum: DatasetDatum) -> str:
        return datum["text"]

    def _get_gold_label_for_datum(self, datum: DatasetDatum) -> str:
        return datum["label"]
