import json
from functools import lru_cache
from pathlib import Path
from typing import Mapping

from datasets import load_dataset

from evoprompt.helpers.prompts import BasePromptsFromJsonMixin
from evoprompt.models import LLMModel
from evoprompt.task import TextClassification
from evoprompt.task.task import DatasetDatum


class AGNews(BasePromptsFromJsonMixin, TextClassification):
    shorthand = "agn"
    base_prompts_file = "evoprompt/initial_prompts/agnews/prompts.json"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(
            *args,
            validation_dataset="fancyzhx/ag_news",
            test_dataset="fancyzhx/ag_news",
            validation_split="train",
            test_split="test",
            **kwargs
        )

    def _get_text_for_datum(self, datum: DatasetDatum) -> str:
        return datum["text"]

    def _get_gold_label_for_datum(self, datum: DatasetDatum) -> str:
        return datum["label"]

    @lru_cache
    def _get_label_mapping(self) -> Mapping:
        classes = ["world", "sports", "business", "tech"]
        return dict(zip(classes, range(len(classes))))


class TREC(BasePromptsFromJsonMixin, TextClassification):
    shorthand = "trec"
    base_prompts_file = "evoprompt/initial_prompts/trec/prompts.json"

    def load_validation_set(
        self, validation_dataset: str | None, validation_split: str | None
    ):
        return load_dataset(
            "csv",
            data_files={"validation": "evoprompt/data/trec/train.csv"},
            split="validation",
            column_names=["label", "question"],
        )

    def load_test_set(self, test_dataset: str, test_split: str | None):
        return load_dataset(
            "csv",
            data_files={"test": "evoprompt/data/trec/test.csv"},
            split="test",
            column_names=["label", "question"],
        )

    def _get_text_for_datum(self, datum: DatasetDatum) -> str:
        return datum["question"]

    def _get_gold_label_for_datum(self, datum: DatasetDatum) -> str:
        return datum["label"]

    @lru_cache
    def _get_label_mapping(self) -> Mapping:
        classes = ["description", "entity", "expression", "human", "location", "number"]
        return dict(zip(classes, range(len(classes))))
